-- A class to allow a developer to display/view debugging output while running the application.
-- See the Notes tab in the DevConsole project for how to use this class.

DevConsole = class()

----------------------------------------------------------------------------------------------------------------------------------
-------------------------- Constants and "static" variables
----------------------------------------------------------------------------------------------------------------------------------

-- Internal Constants

DevConsole.MIN_FONT_SIZE    = 10
DevConsole.MAX_FONT_SIZE    = 18
DevConsole.FONT_INCREMENT   = 1
DevConsole.BUTTON_SIZE      = 28
DevConsole.BUTTON_SPACE     = 20
DevConsole.RESIZE_MARGIN    = 50
DevConsole.TAP_COUNT_DELAY  = 0.1

-- Internal Constants recognized as print controls (and so are NEVER printed, and instead impact :print() behavior)

DevConsole.PRINT_SPACE_SEPARATOR       = {printSeparator=" "}
DevConsole.PRINT_TAB_SEPARATOR         = {printSeparator="\t"}
DevConsole.PRINT_NEWLINE_SEPARATOR     = {printSeparator="\n"}
DevConsole.PRINT_NO_SEPARATOR          = {printSeparator=""}
DevConsole.PRINT_COMMA_SEPARATOR       = {printSeparator=","}
DevConsole.PRINT_COMMA_SPACE_SEPARATOR = {printSeparator=", "} 
DevConsole.PRINT_TEXT_COLOR            = {}
DevConsole.PRINT_HIGHLIGHT             = {}
DevConsole.PRINT_EXPAND_TABLES         = {}
DevConsole.PRINT_COLORS                = {} -- An array that contains all print colors created, each a table with either a color index or actual color
DevConsole.PRINT_SEPARATORS            = {}
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_SPACE_SEPARATOR]                      = DevConsole.PRINT_SPACE_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_TAB_SEPARATOR]                        = DevConsole.PRINT_TAB_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_NEWLINE_SEPARATOR]                    = DevConsole.PRINT_NEWLINE_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_NO_SEPARATOR]                         = DevConsole.PRINT_NO_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_COMMA_SEPARATOR]                      = DevConsole.PRINT_COMMA_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_COMMA_SPACE_SEPARATOR]                = DevConsole.PRINT_COMMA_SPACE_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_SPACE_SEPARATOR.printSeparator]       = DevConsole.PRINT_SPACE_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_TAB_SEPARATOR.printSeparator]         = DevConsole.PRINT_TAB_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_NEWLINE_SEPARATOR.printSeparator]     = DevConsole.PRINT_NEWLINE_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_NO_SEPARATOR.printSeparator]          = DevConsole.PRINT_NO_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_COMMA_SEPARATOR.printSeparator]       = DevConsole.PRINT_COMMA_SEPARATOR
DevConsole.PRINT_SEPARATORS[DevConsole.PRINT_COMMA_SPACE_SEPARATOR.printSeparator] = DevConsole.PRINT_COMMA_SPACE_SEPARATOR

-- Developer Constants: Filter Constants

DevConsole.ANY = {} -- Any filter condition must be met
DevConsole.ALL = {} -- All filter conditions must be met

-- Developer Constants: Default DevConsole Parameter Values

DevConsole.DEFAULT_X                  = 0
DevConsole.DEFAULT_Y                  = 0
DevConsole.DEFAULT_W                  = 1
DevConsole.DEFAULT_H                  = 1

DevConsole.DEFAULT_APPEND_TO_TAB      = false
DevConsole.DEFAULT_BG_COLOR           = color(0,0,0,200)
DevConsole.DEFAULT_BORDER_WIDTH       = 2
DevConsole.DEFAULT_BUFFER_LIMIT       = 100
DevConsole.DEFAULT_COUNT_COLOR        = color(255) 
DevConsole.DEFAULT_ECHO               = false 
DevConsole.DEFAULT_ENABLED            = true 
DevConsole.DEFAULT_FIXED_POSITION     = false
DevConsole.DEFAULT_FIXED_SIZE         = false
DevConsole.DEFAULT_FONT_NAME          = "Inconsolata"
DevConsole.DEFAULT_FONT_SIZE          = 12
DevConsole.DEFAULT_HIDDEN             = false
DevConsole.DEFAULT_HIDE_WHEN_EMPTY    = false
DevConsole.DEFAULT_HIGHLIGHT_COLOR    = color(255,255,0)
DevConsole.DEFAULT_LINE_NUMBERS       = true
DevConsole.DEFAULT_MARGIN             = 20
DevConsole.DEFAULT_MAXIMIZE_ON_PRINT  = false
DevConsole.DEFAULT_MINIMIZE_ACROSS    = false
DevConsole.DEFAULT_MINIMIZE_RIGHT     = true
DevConsole.DEFAULT_MINIMIZE_TOP       = false
DevConsole.DEFAULT_MINIMIZED          = false
DevConsole.DEFAULT_NUMBER_FORMAT      = "%4.3f"
DevConsole.DEFAULT_PRINT_TIME         = true
DevConsole.DEFAULT_PRINT_SEPARATOR    = ""
DevConsole.DEFAULT_SAVE_HIGHLIGHT_CHAR= "•"
DevConsole.DEFAULT_SHOW_ON_PRINT      = true
DevConsole.DEFAULT_TEXT_COLOR         = color(255,255,255,150)
DevConsole.DEFAULT_TIME_FORMAT        = "%M:%S"
DevConsole.DEFAULT_TITLE_SIZE         = 20
DevConsole.DEFAULT_TO_EMPTY_TAB_ONLY  = false
DevConsole.DEFAULT_WRAP_LOG_LINES     = true  

DevConsole.ALT_TIMESTAMP_HMS          = "%T"
DevConsole.ALT_TIMESTAMP_MINSECS      = "%M:%S"
DevConsole.ALT_TIMESTAMP_MICROSECS    = "%f"
DevConsole.ALT_TIMESTAMP_YMDT         = "%y-%m-%d %T"

-- Developer Constants: Parameter Names

DevConsole.PARM_X                        = "x" -- Number: The x position of the console (in pixels or a fraction of the screen width)
DevConsole.PARM_Y                        = "y" -- Number: The y position of the console (in pixels or a fraction of the screen height)
DevConsole.PARM_W                        = "w" -- Number: The width of the console (in pixels or a fraction of the screen width)
DevConsole.PARM_H                        = "h" -- Number: The height of the console (in pixels or a fraction of the screen height)
DevConsole.PARM_MARGIN                   = "margin" -- Number: The size of the margin between the edge of the console and its contents
DevConsole.PARM_TITLE                    = "title" -- String: The title (name) of the console
DevConsole.PARM_TITLE_SIZE               = "titleSize" -- Number: The size of the title of the console
DevConsole.PARM_FONT_NAME                = "fontName" -- String: The font name to use in the console
DevConsole.PARM_FONT_SIZE                = "fontSize" -- Number: The font size to use in the console
DevConsole.PARM_TEXT_COLOR               = "textColor" -- Color: The color of the text printed on the console
DevConsole.PARM_BG_COLOR                 = "bgColor" -- Color: The background color of the console
DevConsole.PARM_TITLE_COLOR              = "titleColor" -- Color: The color of the title of the console
DevConsole.PARM_BORDER_COLOR             = "borderColor" -- Color: The color of the border around the console
DevConsole.PARM_BORDER_WDITH             = "borderWidth" -- Number: The width of the border around the console
DevConsole.PARM_BUFFER_LIMIT             = "bufferLimit" -- Number: The limit to the number of lines in the console buffer (previous lines are discareded)
DevConsole.PARM_HIGHLIGHT_COLOR          = "highlightColor" -- Color: The color used for highlighted printed console lines
DevConsole.PARM_COUNT_COLOR              = "countColor" -- Color: The color used to display the counter for repeated console lines
DevConsole.PARM_TAB_NAME                 = "tabName" -- String: The name of a Codea project tab to which to save the console content (when requested)
DevConsole.PARM_SAVE_HIGHLIGHT_CHAR      = "saveHighlightChar" -- String: The character to use to denote highlighted/colored lines when saving the console to a tab
DevConsole.PARM_FIXED_POSITION           = "fixedPosition" -- Boolean: Prevents the console from being moved by dragging
DevConsole.PARM_FIXED_SIZE               = "fixedSize" -- Boolean: Prevents the console from being resized by dragging
DevConsole.PARM_PRINT_LINE_NUMBERS       = "printLineNumbers" -- Boolean: Prints line numbers on the console
DevConsole.PARM_PRINT_TIME               = "printTime" -- Boolean: Prints the line print time on the console
DevConsole.PARM_PRINT_SEPARATOR          = "printSeparator" -- String: the separator to print between values in a print statement
DevConsole.PARM_ECHO                     = "echo" -- Boolean: Echoes lines printed to the console on the Codea console as well
DevConsole.PARM_ENABLED                  = "enabled" -- Boolean: Enables/disables the console (when disabled, any print statements will be ignored!)
DevConsole.PARM_DISABLED                 = "disabled" -- Boolean: Enables/disables the console (when disabled, any print statements will be ignored!)
DevConsole.PARM_HIDDEN                   = "hidden" -- Boolean: Makes the console hidden
DevConsole.PARM_MINIMIZED                = "minimized" -- Boolean: Makes the console minimized
DevConsole.PARM_MINIMIZE_RIGHT           = "minimizeRight" -- Boolean: Puts the console at the right of the screen when minimized
DevConsole.PARM_MINIMIZE_LEFT            = "minimizeLeft" -- Boolean: Puts the console at the left of the screen when minimized
DevConsole.PARM_MINIMIZE_TOP             = "minimizeTop" -- Boolean: Puts the console at the top of the screen when minimized
DevConsole.PARM_MINIMIZE_BOTTOM          = "minimizeBottom" -- Boolean: Puts the console at the bottom of the screen when minimized
DevConsole.PARM_MINIMIZE_ACROSS          = "minimizeAcross" -- Boolean: Stacks the console horizontally when minimized 
DevConsole.PARM_MINIMIZE_UP_DOWN         = "minimizeUpDown" -- Boolean: Stacks the console vertically when minimized 
DevConsole.PARM_SHOW_ON_PRINT            = "showOnPrint" -- Boolean: Show this console whenever a new line is printed to it
DevConsole.PARM_HIDE_WHEN_EMPTY          = "hideWhenEmpty" -- Boolean: Hide this console when it is empty
DevConsole.PARM_TO_EMPTY_TAB_ONLY        = "toEmptyTabOnly" -- Boolean: Only save the contents, on request, to an empty project tab
DevConsole.PARM_APPEND_TO_TAB            = "appendToTab" -- Boolean: Always append (not replace), on request, console contents to a project tab
DevConsole.PARM_WRAP_LOG_LINES           = "wrapLogLines" -- Boolean: Wrap lines that are too wide to fit in the display
DevConsole.PARM_MAX_ON_PRINT             = "maxOnPrint" -- Boolean: Maximize the console, when minimized, whenever a new line is printed to it
DevConsole.PARM_MINIMIZED_LABEL          = "minimizedLabel" -- String: The one-character label to show when the console is minimized
DevConsole.PARM_TIME_FORMAT              = "timeFormat" -- String: The strformat value to use to format the time for each console line 
DevConsole.PARM_LINE_GAP                 = "lineGap" -- Number: The gap, in pixels, between each console line
DevConsole.PARM_FIRST_LINE               = "firstLine" -- String: An optional line to print as the first line, when the console is created
DevConsole.PARM_PALETTE                  = "palette" -- Table: A list of colors to use for printing by color index
DevConsole.PARM_NUMBER_FORMAT            = "numberFormat" -- The format to use when printing numbers; set to nil or an empty string for no formatting

-- DevConsole Color Themes

DevConsole.THEME_LIGHT 
        = { firstLine="Light Theme", textColor=color(0,200),bgColor=color(220,200), highlightColor=color(0,0,50), borderColor=color(100)}
DevConsole.THEME_DARK 
        = { firstLine="Dark Theme", bgColor=color(0,200),textColor=color(220,200), highlightColor=color(255), borderColor=color(100)}
DevConsole.THEME_NIGHT_SKY 
        = { firstLine="Night Sky Theme", bgColor=color(0,0,20,200),textColor=color(220,200), highlightColor=color(255,255,0,200), borderColor=color(100)}
DevConsole.THEME_DESERT_SANDS 
        = { firstLine="Desert Sands Theme", bgColor=color(120,100,80,200),textColor=color(220,180,140,200), highlightColor=color(0,30,0,220), borderColor=color(240,200,180,200)}
DevConsole.THEME_DRACULA 
        = { firstLine="Dracula Theme", bgColor=color(40,42,54,200),textColor=color(189,147,249),borderColor=color(255,85,85),highlightColor=color(80,250,123)} -- red color(255,85,85) orange color(255,184,108)
DevConsole.THEME_SOLARIZED_BLUE 
        = { firstLine="Solarized Blue Theme", bgColor=color(0,43,54,200),textColor=color(38,139,210),borderColor=color(38,139,210),highlightColor=color(108,113,196),}
DevConsole.THEME_SOLARIZED_GOLD 
        = { firstLine="Solarized Gold Theme", bgColor=color(0,43,54,200),textColor=color(181,137,0),borderColor=color(181,137,0),highlightColor=color(203,75,22),}
DevConsole.THEME_SOLARIZED_LIGHT 
        = { firstLine="Solarized Light Theme", bgColor=color(238,232,213,200),textColor=color(131,107,0),borderColor=color(181,137,0),highlightColor=color(65, 133, 153),}
DevConsole.THEME_SOLARIZED_LIGHTER 
        = { firstLine="Solarized Lighter Theme", bgColor=color(253,246,227,200),textColor=color(181,137,0),borderColor=color(181,137,0),highlightColor=color(65, 133, 153),}
DevConsole.THEME_ZEN_BURN_1 
        = { firstLine="Zen Burn 1 Theme", bgColor=color(143,200),textColor=color(20,66,41),borderColor=color(31),highlightColor=color(23,51,70),}
DevConsole.THEME_ZEN_BURN_2 
        = { firstLine="Zen Burn 2 Theme", bgColor=color(31,200),textColor=color(236,179,179),borderColor=color(223,228,207),highlightColor=color(255,215,167),}
DevConsole.THEME_ZEN_BURN_3 
        = { firstLine="Zen Burn 3 Theme", bgColor=color(31,200),textColor=color(140,176,211),borderColor=color(255,215,167),highlightColor=color(223,228,207),}
DevConsole.THEME_OCEAN 
        = { firstLine="Ocean Theme", bgColor=color(0,50,80,200),textColor=color(0,120,210),borderColor=color(0,50,150),highlightColor=color(140,176,211),}
DevConsole.THEME_FOREST 
        = { firstLine="Forest Theme", bgColor=color(0,80,50,200),textColor=color(180,120,30),borderColor=color(0,150,20),highlightColor=color(0,200,0),}DevConsole.THEME_FLOWERS = {firstLine="Flowers Theme", bgColor=color(10,130,10,200),textColor=color(220,140,30),borderColor=color(160,0,20),highlightColor=color(100,0,200),}
DevConsole.THEME_OUTER_SPACE 
        = {firstLine="Outer Space Theme", bgColor=color(0,220),textColor=color(130),borderColor=color(200),highlightColor=color(180),}
DevConsole.THEME_GRAYISH 
        = {firstLine="Grayish Theme", bgColor=color(60,220),textColor=color(150),borderColor=color(100),highlightColor=color(200),} 
DevConsole.THEME_ROSES 
        = {firstLine="Roses Theme", bgColor=color(100,0,0,220),textColor=color(200,120,120),borderColor=color(255,0,0),highlightColor=color(200,0,0),} 
DevConsole.THEME_FROGS 
        = {firstLine="Frogs Theme", bgColor=color(0,50,0,220),textColor=color(120,200,120),borderColor=color(0,255,0),highlightColor=color(0,200,0),} 
DevConsole.THEME_SUNSET
        = {firstLine="Sunset Theme", bgColor=color(180,100,20,220),textColor=color(120,0,0),borderColor=color(200,200,0),highlightColor=color(200,200,0),}
DevConsole.THEME_STORMY_NIGHT
        = {firstLine="Stormy Night Theme", bgColor=color(20,20,20,220),textColor=color(70,90,140),borderColor=color(130,130,130),highlightColor=color(170,170,200),}
DevConsole.THEME_SPIDERMAN
        = {firstLine="Spiderman Theme", bgColor=color(20,20,220,220),textColor=color(220,0,0),borderColor=color(255),highlightColor=color(0),} 
DevConsole.THEME_SUPERMAN
        = {firstLine="Superman Theme", bgColor=color(20,20,220,220),textColor=color(220,0,0),borderColor=color(180,180,0),highlightColor=color(255,255,0),} 
DevConsole.THEME_WONDER_WOMAN
        = {firstLine="Wonder Woman Theme", bgColor=color(220,20,20,220),textColor=color(180,180,0),borderColor=color(255,255,0),highlightColor=color(255),}  
DevConsole.THEME_BATMAN
        = {firstLine="Batman Theme", bgColor=color(20,240),textColor=color(180,180,0),borderColor=color(95,85,95),highlightColor=color(125,70,145),}
DevConsole.THEME_BLACKBOARD
        = {firstLine="Blackboard", bgColor=color(20,240),textColor=color(200),borderColor=color(240),highlightColor=color(255),fontName="Marker Felt", countColor=color(150)}
DevConsole.THEME_WHITEBOARD
        = {firstLine="Whiteboard", bgColor=color(255,240),textColor=color(20),borderColor=color(20),highlightColor=color(155,0,0),fontName="Noteworthy", countColor=color(0,0,255)} 

-- Zen Reference Colors (delete this)
-- pale mauve 204,147,147, pale forest 95,127,95, pale peach 255,215,167, pale sky 140,176,211
-- aquamarine 113,211,180, green ice 223,228,207, pale salmon 236,179,179, palish peach 255,207,175
-- darkest gray 31, middle gray 143, 


-- Icon Consoles for Buttons (UTF-8 Symbols)

DevConsole.ICON_DOWN_ARROW_TO_BAR = "\xE2\xA4\x93"
DevConsole.ICON_DOWN_ARROW_FROM_BAR = "\xE2\x86\xA7"
DevConsole.ICON_DOWN_WHITE_ARROW = "\xE2\x87\xA9"
DevConsole.ICON_CLOSED_LOCK = "\xF0\x9F\x94\x92"
DevConsole.ICON_OPEN_LOCK = "\xF0\x9F\x94\x93"
DevConsole.ICON_BULLSEYE = "\xE2\x97\x8E"
DevConsole.ICON_WHITE_SCISSORS = "\xE2\x9C\x84"
DevConsole.ICON_UPPER_BLADE_SCISSORS = "\xE2\x9C\x81"
DevConsole.ICON_LOWER_RIGHT_PENCIL = "\xE2\x9C\x8E"
DevConsole.ICON_NEXT_PAGE = "\xE2\x8E\x98"
DevConsole.ICON_CLEAR_SCREEN = "\xE2\x8E\x9A"
DevConsole.ICON_PRINT_SCREEN = "\xE2\x8E\x99"
DevConsole.ICON_ARROW_IN_BOX = "\xE2\x8D\x97" -- APL Functional Symbol Quad Downwards Arrow
DevConsole.ICON_STAR_FILLED = "\xE2\x98\x85" -- ★
DevConsole.ICON_STAR_OUTLINE = "\xE2\x98\x86" -- ☆ 
DevConsole.ICON_STAR_EMOJI = "\xE2\xAD\x90" --  ⭐
DevConsole.ICON_LIGHTNING_EMOJI = "\xE2\x9A]xA1" -- ⚡ 
DevConsole.ICON_POINTY_STAR = "xE2\x9C\xB9" -- ✹
DevConsole.ICON_ALERT = "\xE2\x9A\xA0" -- ⚠
                    
-- Static variables: Internal Use Only!

DevConsole.CONSOLES = {}
--setmetatable(DevConsole.CONSOLES, {__mode="v", __name="DevConsole"})

-- tlv = top left vertical, brh = bottom right horizontal
DevConsole.MINIMIZED = {tlv={},blv={},trv={},brv={},tlh={},blh={},trh={},brh={}}

----------------------------------------------------------------------------------------------------------------------------------
-------------------------- Static methods: For Developer Use and Integration into Main setup(), Main draw() and Main touched()
----------------------------------------------------------------------------------------------------------------------------------

-- Saved copy of the Codea print function
DevConsole.luaPrint = print -- Save this for later use in case a developer overrides it to write to a console

function DevConsole.createPrintColor(printColor, extra)
    -- This trickery lets a developer call this method using .createPrintColor() or :createPrintColor()
    if type(printColor) == "table" and printColor.is_a and printColor:is_a(DevConsole) then
        printColor = extra
    end
    if type(printColor) == "number" or type(printColor) == "userdata" and printColor.r and printColor.g and printColor.b then
        if DevConsole.PRINT_COLORS[printColor] then return DevConsole.PRINT_COLORS[printColor] end
        local result = {printColor=printColor}
        DevConsole.PRINT_COLORS[printColor] = result -- This version makes it easy to find the setting for a color
        DevConsole.PRINT_COLORS[result] = result -- This version makes it easy to find the print setting
        return result
    else
        error("Invalid createPrintColor parameter value " .. tostring(printColor))
    end
end

function DevConsole.createPrintSeparator(printSeparator, extra)
    -- This trickery lets a developer call this method using .createPrintSeparator() or :createPrintSeparator()
    if type(printSeparator) == "table" and printSeparator.is_a and printSeparator:is_a(DevConsole) then
        printSeparator = extra
    end
    if type(printSeparator) == "string" then
        if DevConsole.PRINT_SEPARATORS[printSeparator] then return DevConsole.PRINT_SEPARATORS[printSeparator] end
        local result = {printSeparator=printSeparator}
        DevConsole.PRINT_SEPARATORS[printSeparator] = result -- This version makes it easy to find the setting for a separator
        DevConsole.PRINT_SEPARATORS[result] = result -- This version makes it easy to find the print setting
        return result
    else
        error("Invalid createPrintSeparator parameter value " .. tostring(printColor))
    end
end

-- "Static" method to let all consoles handle touches 
DevConsole.touchedAny = function(touch)
    -- Let any console that already has that touch get the update
    for i,console in ipairs(DevConsole.CONSOLES) do
        if console.touches[touch.id] then
            if console:touched(touch) then return true end
        end
    end
    -- Let mimimized consoles get touches first
    for k,set in pairs(DevConsole.MINIMIZED) do
        for i,console in ipairs(set) do
            if console:touched(touch) then return true end
        end
    end
    -- Offer the touch to each console in reverse order
    for i=#DevConsole.CONSOLES,1,-1 do
         if DevConsole.CONSOLES[i]:touched(touch) then return true end -- If this console handles the touch, then don't let anyone else
    end
    return false
end

-- "Static" method to let 
function DevConsole.drawAll()
    pushMatrix()
    pushStyle()
    resetMatrix()
    resetStyle()
    for i,console in ipairs(DevConsole.CONSOLES) do
        if not console.minimized then console:draw() end
    end
    for i,corner in pairs(DevConsole.MINIMIZED) do
        for j,console in ipairs(corner) do 
            console:draw()
        end
    end
    popStyle()
    popMatrix()
end

-- Restore all consoles to their original locations and sizes and show them
function DevConsole.restoreAll()
    for i,console in ipairs(DevConsole.CONSOLES) do
        console:restoreToOrigin()
        console:maximize() -- show and maximize
    end
end

-- Get a specific console, by index (a number) or a pattern (a string) applied to the console title
function DevConsole.getConsole(parm)
    if type(parm) == "number" then
        if parm <= 0 then parm = #DevConsole.CONSOLES - parm end
        return DevConsole.CONSOLES[parm] 
    end
    assert(type(parm) == "string", "Use a number (index) or string (pattern) as the parameter to DevConsole.getConsole()")
    for i,console in ipairs(DevConsole.CONSOLES) do
        if string.match(console.title, parm) then return console end
    end
end

-- Call this in the Main draw() method INSTEAD of .drawALl() to generate tests
-- consoles: The number of consoles to create
-- parms   : Any console parameters to use; specify a table of options to choose randomly from a set of option values
function DevConsole.testAll(consoles, commonParms, delay)
    
    -- Turn off testing (destroying test consoles) by passing "false" for consoles
    
    if consoles == false then
        DevConsole.testDevConsoles = nil
        return
    end
    
    if type(consoles) ~= "number" then consoles = 5 end -- Default number of consoles is 5
    
    -- Create console data if this is the first test call
    
    if DevConsole.testDevConsoles == nil then
        -- Just some silly words used to create random log statements and titles
        DevConsole.testWords = {
            {"The","Any","Many","Few","No","All","Those","These"},
            {"Big,","Small,","Tiny,","Huge,","Smelly,","Stinky,","Ugly,","Angry,","Mopey,","Curious,"},
            {"Blue","Red","Green","Loud","Quiet","Grumpy","Sad","Tired","Lazy"},
            {"Bats","Rats","Cats","Dogs", "Bugs", "Eels", "Pigs", "Hogs","Frogs","Gnats","Slugs"},
            {"Hate","Love","Eat","Drink","Devour","Inhale"},
            {"Poisonous","Foul","Malodorous","Insidious","Moldy","Slimy","Runny","Useless","Venomous"},
            {"Smoke","Flowers","Juices","Puddles","Spaghetti","Cheese","Rocks","Feathers","Air","Girls","Boys","Puppies","Kittens"}
        }
        DevConsole.testButs = {"And", "If", "But", "Or", "Even Though", "Whenever", "But Maybe"}
        
        -- Track the last time something was written to one of the consoles
        DevConsole.lastTestPrint = os.clock()
        -- Make a persistent list of the test consoles created (so they aren't garbage collected)
        DevConsole.testDevConsoles = {}
        DevConsole.testDevConsoleCount = 0
    
        while DevConsole.testDevConsoleCount < consoles do
            commonParms = commonParms or {
                showOnPrint=false, 
                bufferLimit=100, 
                title=DevConsole.testWords[4],
                textColor={color(255,0,0),color(0,255,0),color(0,255,255),color(255,0,255)}
            }
            -- Construct default parms from parameter passed, choosing randomly when more than one option
            local parms = {}
            for k,v in pairs(commonParms) do
                -- If the value is a table with numeric values, randomly select a value from the table
                if type(v) == "table" and #v > 0 then v = v[math.random(#v)] 
                elseif type(v) == "function" then v = v()
                end
                parms[k] = v
            end
            -- Make sure the console has a location, specified or random
            parms.x = parms.x or math.random() * 0.6
            parms.y = parms.y or math.random() * 0.6
            parms.w = parms.w or math.min(math.max(0.4,math.random()),1-parms.x)
            parms.h = parms.h or math.min(math.max(0.4,math.random()),1-parms.y)
            -- Create and remember the actual console 
            table.insert(DevConsole.testDevConsoles, DevConsole(parms))
            DevConsole.testDevConsoleCount = DevConsole.testDevConsoleCount + 1
        end
    end
    
    -- Make a random log entry into one of the logs
    
    if DevConsole.testDevConsoleCount > 0 and os.clock() > DevConsole.lastTestPrint + (delay or 0) then
        local index = math.random(DevConsole.testDevConsoleCount)
        local phrase = ""
        for i,words in ipairs(DevConsole.testWords) do
            phrase = phrase .. words[math.random(#words)] .. " "
        end
        if math.random() > 0.8 then
            phrase = phrase .. DevConsole.testButs[math.random(#DevConsole.testButs)] .. " "
            for j,more in ipairs(DevConsole.testWords) do
                phrase = phrase .. more[math.random(#more)] .. " "
            end
            DevConsole.testDevConsoles[index]:printWithColor(true,phrase)
        else
            DevConsole.testDevConsoles[index]:print(phrase)
        end
        DevConsole.lastTestPrint = os.clock()
    end
    
    -- Do the draw all here (so the developer isn't confused if he forgets to include it)
    
    DevConsole.drawAll()
end

---
--- ===== DevConsole Class Methods ===== ---
---

----------------------------------------------------------------------------------------------------------------------------------
-------------------------- Constructor
----------------------------------------------------------------------------------------------------------------------------------

-- Initialize this console
function DevConsole:init(...)
    DevConsole.preventGc = self
    
    -- Separate parameters into lists of numbers, colors, and strings, to be used in order where appropriate
    
    local numbers,colors,strings,booleans = {}, {}, {}, {}
    local numberIndex, colorIndex, stringIndex, booleanIndex = 0,0,0,0
    local parms
    if select("#",... ) == 1 and type(select(1,...)) == "table" then parms = select(1,...) else parms = table.pack(...) end
    local extractParms
    extractParms = function(inparms)
        for i,v in ipairs(inparms) do
            if type(v) == "number" then table.insert(numbers, v)
            elseif type(v) == "string" then table.insert(strings, v)
            elseif type(v) == "userdata" and v.r and v.g and v.b then table.insert(colors, v)
            elseif type(v) == "table" then 
                extractParms(v)
                for k,v2 in pairs(v) do
                    if type(k) == "string" and parms[k] == nil then
                        parms[k] = v2
                    end
                end
            end
        end
    end
    extractParms(parms)    
        
    -- Internal, one-time functions for extracting the right init parameters
    
    -- Return the next number in the parameter list (nil if no more)
    local nextNumber = function(name, default)
        assert(parms[name] == nil or type(parms[name]) == "number", 
                "Invalid (non-number) value for console parameter " .. tostring(name) )
        if parms[name] then return parms[name] end
        numberIndex = numberIndex + 1
        return numbers[numberIndex] or default
    end
    
    -- Return the next color in the parameter list (nil if no more)
    local nextColor = function(name, default)
        assert(parms[name] == nil or (type(parms[name]) == "userdata" and parms[name].r and parms[name].g and parms[name].b), 
                "Invalid (non-color) value for console parameter " .. tostring(name) )
        if parms[name] then return parms[name] end
        colorIndex = colorIndex + 1
        return colors[colorIndex] or default
    end
    
    -- Return the next string in the parameter list (nil if no more)
    local nextString = function(name, default)
        assert(parms[name] == nil or type(parms[name]) == "string", 
                "Invalid (non-string) value for console parameter " .. tostring(name) )
        if parms[name] then return parms[name] end
        stringIndex = stringIndex + 1
        return strings[stringIndex] or default
    end
    
    -- Return the next boolean in the parameter list (nil if no more)
    local nextBoolean = function(name, default, opposite)
        assert(parms[name] == nil or type(parms[name]) == "boolean", 
                "Invalid (non-boolean) value for console parameter " .. tostring(name) )
        if parms[name] ~= nil then return parms[name] end
        if opposite ~= nil then
            assert(parms[opposite] == nil or type(parms[opposite]) == "boolean", 
                    "Invalid (non-boolean) value for console parameter " .. tostring(opposite) )
            if parms[opposite] ~= nil then return not parms[opposite] end
        end
        booleanIndex = booleanIndex + 1
        if booleans[booleanIndex] ~= nil then return booleans[booleanIndex] end
        return default
    end
    
    ------------------------------------------------------------------------------------------------------------------------
    
    -- Grab all parameters
    
    -- Location of the DevConsole in screen coordinates
    
    self.x           = nextNumber(DevConsole.PARM_X , DevConsole.DEFAULT_X) -- The Y location of the lower left corner of the console
    self.y           = nextNumber(DevConsole.PARM_Y , DevConsole.DEFAULT_Y) -- The X location of the lower left corner of the console
    self.w           = nextNumber(DevConsole.PARM_W , DevConsole.DEFAULT_W) -- Defaults to current WIDTH when drawn
    self.h           = nextNumber(DevConsole.PARM_H , DevConsole.DEFAULT_H) -- Defaults to current HEIGHT when drawn
    self.margin      = nextNumber(DevConsole.PARM_MARGIN , DevConsole.DEFAULT_MARGIN) -- The width of an empty inner margin between the edge of the console and its contents
    
    -- Remember original location and dimensions
    self.origin      = {x=self.x,y=self.y,w=self.w,h=self.h}
    
    -- Set title
    self.title       = nextString(DevConsole.PARM_TITLE , "") -- The title to display on the console
    self.titleSize   = nextNumber(DevConsole.PARM_TITLE_SIZE , DevConsole.DEFAULT_TITLE_SIZE) -- The size of the title font
    
    self.fontName    = nextString(DevConsole.PARM_FONT_NAME , DevConsole.DEFAULT_FONT_NAME) -- The font for the console
    self.fontSize    = nextNumber(DevConsole.PARM_FONT_SIZE , DevConsole.DEFAULT_FONT_SIZE) -- The default size of the console text font
    self.resetFontSize= self.fontSize -- The base font size (for reseting the console display)
    
    -- Colors
    
    self.textColor   = nextColor(DevConsole.PARM_TEXT_COLOR , DevConsole.DEFAULT_TEXT_COLOR) -- The color of the text on the console
    self.bgColor     = nextColor(DevConsole.PARM_BG_COLOR , DevConsole.DEFAULT_BG_COLOR) -- The background color of the console
    self.titleColor  = nextColor(DevConsole.PARM_TITLE_COLOR , self.textColor) -- The color of the title
    self.borderColor = nextColor(DevConsole.PARM_BORDER_COLOR , self.titleColor) -- The color of the border (frame) around the console
    self.borderWidth = nextNumber(DevConsole.PARM_BORDER_WDITH , DevConsole.DEFAULT_BORDER_WIDTH) -- The width of the border (frame) around the console
    self.bufferLimit = nextNumber(DevConsole.PARM_BUFFER_LIMIT , DevConsole.DEFAULT_BUFFER_LIMIT) -- The limit to the number of lines in the buffer. <=0 means no limit.
    self.highlightColor = nextColor(DevConsole.PARM_HIGHLIGHT_COLOR , DevConsole.DEFAULT_HIGHLIGHT_COLOR) -- The color of highlighted print statements
    self.countColor  = nextColor(DevConsole.PARM_COUNT_COLOR , DevConsole.DEFAULT_COUNT_COLOR)
    
    -- Settings
    
    self.tabName       = nextString(DevConsole.PARM_TAB_NAME , string.gsub(self.title, "^[^%w]+", "_")) -- The name of the tab to which to write the console
    self.saveHighlightChar = nextString(DevConsole.PARM_SAVE_HIGHLIGHT_CHAR, DevConsole.DEFAULT_SAVE_HIGHLIGHT_CHAR) -- The character to denote a hilighted/colored line in saved contents
    self.fixedPosition = nextBoolean(DevConsole.PARM_FIXED_POSITION , DevConsole.DEFAULT_FIXED_POSITION) -- Can the console be moved on the screen by the user
    self.fixedSize     = nextBoolean(DevConsole.PARM_FIXED_SIZE , DevConsole.DEFAULT_FIXED_SIZE) -- Can the console be resized by the user
    self.printLineNumbers   = nextBoolean(DevConsole.PARM_PRINT_LINE_NUMBERS , DevConsole.DEFAULT_LINE_NUMBERS) 
    self.printTime     = nextBoolean(DevConsole.PARM_PRINT_TIME , DevConsole.DEFAULT_PRINT_TIME) 
    self.printSeparator= nextString(DevConsole.PARM_PRINT_SEPARATOR, DevConsole.DEFAULT_PRINT_SEPARATOR)
    self.enabled       = nextBoolean(DevConsole.PARM_ENABLED , DevConsole.DEFAULT_ENABLED, DevConsole.PARM_DISABLED) -- If disabled, print statements are ignored
    self.echo          = nextBoolean(DevConsole.PARM_ECHO , DevConsole.DEFAULT_ECHO) -- Should print statements echo on the Codea console
    self.hidden        = nextBoolean(DevConsole.PARM_HIDDEN , DevConsole.DEFAULT_HIDDEN) -- Is the console initially hidden?
    self.minimized     = nextBoolean(DevConsole.PARM_MINIMIZED , DevConsole.DEFAULT_MINIMIZED) -- Is the console initially minimized?
    self.minimizeRight = nextBoolean(DevConsole.PARM_MINIMIZE_RIGHT , DevConsole.DEFAULT_MINIMIZE_RIGHT , DevConsole.PARM_MINIMIZE_LEFT) -- When minimized, push the console marker to the left side of the display
    self.minimizeTop   = nextBoolean(DevConsole.PARM_MINIMIZE_TOP , DevConsole.DEFAULT_MINIMIZE_TOP , DevConsole.PARM_MINIMIZE_BOTTOM) -- When minimized, push the console marker to the top of the display
    self.minimizeAcross= nextBoolean(DevConsole.PARM_MINIMIZE_ACROSS , DevConsole.DEFAULT_MINIMIZE_ACROSS, DevConsole.PARM_MINIMIZE_UP_DOWN) -- When minimized, stack horizontally (across)
    self.showOnPrint   = nextBoolean(DevConsole.PARM_SHOW_ON_PRINT , DevConsole.DEFAULT_SHOW_ON_PRINT) -- Show the console whenever a new line is printed
    self.hideWhenEmpty = nextBoolean(DevConsole.PARM_HIDE_WHEN_EMPTY , DevConsole.DEFAULT_HIDE_WHEN_EMPTY) -- Hide the console whenever empty
    self.toEmptyTabOnly    = nextBoolean(DevConsole.PARM_TO_EMPTY_TAB_ONLY , DevConsole.DEFAULT_TO_EMPTY_TAB_ONLY) -- If the target save tab is not empty, create a new one
    self.appendToTab   = nextBoolean(DevConsole.PARM_APPEND_TO_TAB , DevConsole.DEFAULT_APPEND_TO_TAB) -- Append to (don't overwrite) a tab
    self.wrapLogLines  = nextBoolean(DevConsole.PARM_WRAP_LOG_LINES , DevConsole.DEFAULT_WRAP_LOG_LINES) -- Wrap log lines when writing to the screen
    self.maxOnPrint  = nextBoolean(DevConsole.PARM_MAX_ON_PRINT , DevConsole.DEFAULT_MAXIMIZE_ON_PRINT) -- De-minimize on print
    
    self.minimizedLabel = nextString(DevConsole.PARM_MINIMIZED_LABEL) -- The one-character label to use for the console in minimized form
    self.timeFormat    = nextString(DevConsole.PARM_TIME_FORMAT , DevConsole.DEFAULT_TIME_FORMAT) -- The strftime time format to use when printing times for logged print lines
    self.lineGap       = nextNumber(DevConsole.PARM_LINE_GAP) -- Gap between lines on the console (useful for distinguishing wrapped text)
    if self.lineGap == nil then self.lineGap = self.fontSize / 3 end
    self.firstLine     = nextString(DevConsole.PARM_FIRST_LINE) -- The first line to display on the console
    self.numberFormat  = nextString(DevConsole.PARM_NUMBER_FORMAT, DevConsole.DEFAULT_NUMBER_FORMAT) -- The number format to use whenever printing numbers
    
    self.palette = parms[DevConsole.PARM_PALETTE]
    if self.palette == nil then self:createPalette() end
    
    -- Internal storage
    
    self.origin = {x=self.x,y=self.y,w=self.w,h=self.h}
    self.nextLine= 1 -- The next line number 
    self.offset  = vec2() -- This is the scrolling offset, to shift the content 
    self.contentSize    = vec2() -- This is the computed size of the entire content in pixels, after being drawn (for scrolling/zoom purposes)
    self.buffer  = {} -- This is the content of the console; contents are tables of strings and colors as {str=, col=}
    self.colorFilters = {}
    self.savedTabs = {}
    
    if type(self.tabName) == "string" and string.len(self.tabName) > 0 and not self.toEmptyTabOnly and not self.appendToTab then
        self:printWithColor(true, "════════ ════════ ════════ ════════ ════════ ")
        self:printWithColor(true, "WARNING: DevConsole saves to ",self.tabName," will overwrite any existing project data in that tab")
        self:printWithColor(true, "════════ ════════ ════════ ════════ ════════ ")
    end
    
    if self.firstLine then self:print(self.firstLine) end
    
    -- Create buttons
    
    self.buttons = {}
    self.touches = {}
    self.touchCount = 0
    
    local createButton = function(label, action, delay, atLeft, atBottom)
        table.insert(self.buttons, {
            label=label,action=action,
            atLeft=atLeft or false,atBottom=atBottom or false,
            delay=delay or 0})
        return self.buttons[#self.buttons]
    end
    createButton("◎", function() self:zoomReset() end, 0, true)
    createButton("+", function() self:zoomIn() end, 0.1, true)
    createButton("-", function() self:zoomOut() end, 0.1, true)
    createButton("⎘", function() self:copyToPasteboard() end, 0, true)
    
    createButton("X", function() self:minimize() end) -- ❌
    if type(self.tabName) == "string" and string.len(self.tabName) > 0 then
        createButton("⇩", function() self:saveToTab() end) -- or "S" or "⇭" or "⇩" or "⇩" or "✎" ✏
    end
    createButton("✄", function() self:clear() end) -- ✄ 🧹
    self.lockButton = createButton("🔒", function() if self.locked then self:unlock() else self:lock() end end) -- ⎘
    
    
    -- Give it a minimize label
    if self.minimizedLabel == nil then
        if self.title and string.len(self.title) > 0 then
            self.minimizedLabel = string.sub(self.title,1,1)
        else
            self.title = "Console " .. tostring(#DevConsole.CONSOLES + 1)
            if self.tabName == nil or string.len(self.tabName) == 0 then
                self.tabName =string.gsub(self.title, "%s", "")
            end
            self.minimizedLabel = tostring(#DevConsole.CONSOLES)
        end
    elseif self.title == nil or string.len(self.title) == 0 then 
        self.title = "Console " .. tostring(#DevConsole.CONSOLES + 1)
    end
    
    -- Add the console to the list of known consoles
    
    table.insert(DevConsole.CONSOLES, self)
    
    if self.minimized then self:minimize() end
end

function DevConsole:close()
    for i,console in ipairs(DevConsole.CONSOLES) do
        if self == console then
            table.remove(DevConsole.CONSOLES,i)
            return
        end
    end
end

----------------------------------------------------------------------------------------------------------------------------------
-------------------------- Behaviors and Control and  Utility Methods
----------------------------------------------------------------------------------------------------------------------------------

-- Clear the buffer for this console (i.e., erase its contents)
function DevConsole:clear()
    self.buffer = {}
    self.priorLine = nil
    self.nextLine = 1
    if self.hideWhenEmpty then
        self:hide()
    end
end

-- Get the current buffer as a string
-- Optionally, specify a range of line numbers to include
-- Optionally, override the printTime and line number options
function DevConsole:getAsString(fromLine, toLine, printTime, printLineNumbers)
    
    if printTime == nil then printTime = self.printTime end
    if printLineNumbers == nil then printLineNumbers = self.printLineNumbers end
    
    newData = "--[[\n"
    newData = newData .. string.rep("━",60) .. "\n"
    newData = newData .. "Console Output for " .. self.title .. " from " .. os.date("%c",os.time()) .. "\n"
    newData = newData .. string.rep("─",60) .. "\n"
    
    -- TODO: Apply display filter, or offer a separate saveOrCopyFilter?
    
    -- Contents
    
    local lineStr 
    local lineStart
    for i,bufferLine in ipairs(self.buffer) do
        if fromLine == nil or bufferLine.lin >= fromLine then
            if toLine == nil or bufferLine.lin <= toLine then
                lineStr = "--\t"
                if bufferLine.col == self.textColor then 
                    lineStr = lineStr .. "\t"
                else
                    lineStr = lineStr ..  (self.saveHighlightChar or "•") .. "\t" 
                end
                if printTime then
                    lineStr = lineStr .. os.date(" %y %m %d %H:%M:%S\t",bufferLine.tim)
                end
                if printLineNumbers then
                    lineStr = lineStr .. string.format("%04d\t",bufferLine.lin)
                end
                lineStr = lineStr .. string.gsub(bufferLine.str,"%]%]","]\\]")
                if bufferLine.cnt > 1 then
                    lineStr = lineStr .. "\t\xF0\x9F\xA4\x87" .. tostring(bufferLine.cnt) .. "\xE2\xA8\xB5	 until " .. os.date("%H:%M:%S", bufferLine.lst)
                end
                newData = newData .. lineStr .. "\n"
            end
        end
    end
    
    -- TODO: Option to include pending lines when console is locked
    
    -- Footer
    
    newData = newData .. "\n" .. string.rep("━",60) .. "\n\n"
    newData = newData .. "--]]"
    
    return newData
    
end

-- Restore the console to its original location and size (for use after the console has been dragged or resized)
function DevConsole:restoreToOrigin()
    self.x = self.origin.x
    self.y = self.origin.y
    self.w = self.origin.w
    self.h = self.origin.h
end

-- Copy the contents of the console to the pastebooard
function DevConsole:copyToPasteboard(...)
    pasteboard.text = self:getAsString(...)
    self:printWithColor(true,"Log contents copied to pasteboard")
end

function DevConsole:lock()
    self.locked = true
    if self.pendingLines == nil then self.pendingLines = {} end
    self.lockButton.label = "🔓"
end

function DevConsole:unlock()
    self.locked = false
    self.lockButton.label = "🔒"
    -- Transfer buffered lines
    local pastPending
    pastPending, self.pendingLines = self.pendingLines, {}
    for i,pendingLine in ipairs(self.pendingLines) do
        self.priorLine = pendingLine
        table.insert(self.buffer, pendingLine)
    end
    self:afterPrint()
end

-- Add some of the console settings to the console itself
function DevConsole:printSettings()
    self:print("title:",self.title)
    self:print("x:",self.x,", y:",self.y,", w:",self.w,", h:",self.h)
    for k,v in pairs(self) do
        if type(k) == "string" then
            if type(v) == "number" or type(v) == "string" or type(v) == "boolean" or type(v) == "userdata" then
                self:printWithColor(true, k," = ",v)
            end
        end
    end
    --self:print("border: ",self.borderWidth," ",self.borderColor)
    --self:print("bg: ",self.bgColor,", text:",self.textColor)
    --self:print()
end

function DevConsole:createPalette(textColor, highlightColor, bgColor, fudge)
    
    textColor = textColor or self.textColor or DevConsole.DEFAULT_TEXT_COLOR
    highlightColor = highlightColor or self.highlightColor or DevConsole.DEFAULT_HIGHLIGHT_COLOR
    bgColor = bgColor or self.bgColor or DevConsole.DEFAULT_BG_COLOR
    fudge = fudge or 32
    fudgeHalf = math.floor(fudge / 2)
    
    -- Create the palette with text color as #1 and highlight color as #2, even if they are the same
    
    self.palette = {}
    
    local fudged = function(v)
        return math.min(255, math.floor(v / fudge) * fudge + fudgeHalf)
    end

    -- Temp table to track colors used already
    
    local usedTable = {}
    local doNotAdd = function(r,g,b,a)
        
        -- Don't add if this exact color exists
        if usedTable[r] and usedTable[r][g] and usedTable[r][g][b] then return true end
        if not usedTable[r] then usedTable[r] = {} end
        if not usedTable[r][g] then usedTable[r][g] = {} end
        usedTable[r][g][b] = true
        
        -- Don't add if a fudge for this color exists
        r,g,b = fudged(r), fudged(g), fudged(b)
        if usedTable[r] and usedTable[r][g] and usedTable[r][g][b] then return true end
        -- Gonna add, so remember it for later
        if not usedTable[r] then usedTable[r] = {} end
        if not usedTable[r][g] then usedTable[r][g] = {} end
        usedTable[r][g][b] = true
        
        return false
    end

    -- Function to add colors without duplicates
        
    local addColor = function(r,g,b,a)
        local newColor = color(r,g,b,a)
        
        if newColor == bgColor then return end -- Never add the background color
        if doNotAdd(r,g,b,a) then return end -- no duplicates allowed
        
        table.insert(self.palette, newColor)
    end
    
    addColor(textColor.r, textColor.g, textColor.b, textColor.a)
    addColor(highlightColor.r, highlightColor.g, highlightColor.b, highlightColor.a)
    doNotAdd(bgColor.r, bgColor.g, bgColor.b, bgColor.a)
    -- 1 and 2 are always text color and highlight color, even if they're the same
    if #self.palette == 1 then table.insert(self.palette, highlightColor) end
    
    -- Function to generate variations of a color and put them in the destination table
    
    local permutations = function(t,a)
        if a == 0 then return end
        local r,g,b
        
        for i=1,3 do
            r=t[i]
            g=t[(i+0)%3+1]
            b=t[(i+1)%3+1]
            addColor(r,g,b,a)
            addColor(r,b,g,a)
            addColor(r,g,g,a)
            addColor(r,b,b,a)
        end 
    end
    
    local variations = function(baseColor)
        -- First, do all combinations or r,g,b
        local r,g,b,a = baseColor.r, baseColor.g, baseColor.b, baseColor.a
        
        permutations({r, g, b}, a)
        
        local hi, lo = math.max(r,g,b), math.min(r,g,b) 
        
        local swap = function(a,b,c) if a == b then return c else return a end end 
        
        --  Next, try repeating the highest value in the lowest spot
        --permutations(swap(r,lo,hi),swap(g,lo,hi),swap(b,lo,hi),a)
        
        --  Next, try changing the largest value to 255
        if #self.palette < 8 then
            permutations({swap(r,hi,255),swap(g,hi,255),swap(b,hi,255)},a)
        end
        
        --  Next, try changing the lowest value to 0
        if #self.palette < 8 then
            permutations({swap(r,lo,0),swap(g,lo,0),swap(b,lo,0)},a)
        end
        
        -- Last, try intense colors
        if #self.palette < 8 then
            permutations({255,128,0}, a)
            permutations({255,0,0}, a)
        end
    end
    
    variations(textColor)
    variations(highlightColor)
    
end
    
function DevConsole:getPalette()
    return self.palette
end

function DevConsole:setPalette(palette)
    self.palette = palette
end

function DevConsole:getPaletteSize()
    return #self.palette
end

function DevConsole:logPalette()
    for i,c in ipairs(self.palette) do
        self:printWithColor(i,i,":",c)
    end
end

function DevConsole:enable()
    self:setEnabled(true)
end

function DevConsole:disable()
    self:setEnabled(false)
end

function DevConsole:setEnabled(enabled)
    if not enabled then self:printc(true,"Console disabled!") end
    self.enabled = enabled or false
end

function DevConsole:getFilterFunction(...)
    
    local passedParms = table.pack(...)
    
    -- Clone the parameter table (recursively), so it won't be changed, creating unexpected side effects
    -- While doing so, validate that the parameters are all well-formatted
    
    local any = true
    
    local parms = {}
    local function packParms(testParms) 
        for i,parm in ipairs(testParms) do
            if parm == DevConsole.ANY then
                any = true
            elseif parm == DevConsole.ALL then
                any = false
            elseif type(parm) == "table" then
                if #parm == 2 and type(parm[1]) == "number" and type(parm[2]) == "number" then
                    -- Number pairs must be the same time, i.e. lines or times
                    assert(parm[1] > 9999999 and parm[2] > 9999999 or parm[1] <= 9999999 and parm[2] <= 9999999,
                        "Pairs of filter numbers must be in the same range (i.e. <=9999999 or >9999999)")
                    table.insert(parms, parm)
                else
                    packParms(parm)
                end
            elseif type(parm) == "userdata" then
                assert(parm.r and parm.g and parm.b and parm.a,"The only user data accepted by a filter is a color")
                table.insert(parms, parm) 
            elseif parm ~= nil then 
                table.insert(parms, parm) 
            end
        end
    end
    
    packParms(passedParms)
    
    -- No result if no filter parameters
    
    if #parms == 0 then return nil end
    
    -- Create the resulting filter function to use
    
    return function(bufferLine)
        if bufferLine == nil then return false end
        local parmType, parmResult
        for i,parm in ipairs(parms) do
            parmType = type(parm)
            parmResult = false
            if parmType == "string" then
                if string.match(bufferLine.str, parm) then parmResult = true end
            elseif parmType == "number" then
                -- time
                if parm > 9999999 then -- Time
                    if bufferLine.tim >= parm then parmResult = true end
                else -- Line number
                    if bufferLine.lin >= parm then parmResult = true end
                end
            elseif parmType == "function" then 
                if parm(bufferLine) then parmResult = true end
            elseif parmType == "boolean" then
                if parm and bufferLine.col == self.highlightColor then parmResult = true end
                if not parm and bufferLine.col ~= self.highlightColor then parmResult = true end
            elseif parmType == "userdata" then -- This has been checked to be a color
                if bufferLine.col == parm then parmResult = true end
            elseif parmType == "table" then -- This as been checked to be a pair of valid numbers only
                if parm[1] > 9999999 then -- Time
                    -- ANY overlap counts (times are ranges, so it's not a simple comparison)
                    if not ( bufferLine.lst < parm[1] or bufferLine.tim > parm[2] ) then parmResult = true end
                else
                    if bufferLine.lin >= parm[1] and bufferLine.lin <= parm[2] then parmResult = true end
                end
            end
            
            if any == parmResult then return parmResult end
        end
        
        return not any
    end
end

function DevConsole:clearColorFilters()
    self.colorFilters = {}
end

function DevConsole:setColorFilter(lineColor, ...)
    if lineColor == nil  and select("#",...) == 0 then
        self.colorFilters = {}; return
    elseif lineColor == true then
        lineColor = self.highlightColor
    else
        assert(type(lineColor) == "userdata" and lineColor.r and lineColor.g and lineColor.b,
        "The first value passed to DevConsole:setColorFilter must be a color (or nil)")
    end

    -- If no filter parameters, then remove any filter for this color
        
    if select("#",...) == 0 then
        for i,v in ipairs(self.colorFilters) do
            if v.lineColor == lineColor then
                table.remove(self.colorFilters, i)
                break
            end
        end
        return
    end
    
    -- Create the filter function
    
    local filterFunc = self:getFilterFunction(...)
    
    -- If the color already exists, change the filter function
    
    for i,v in ipairs(self.colorFilters) do
        if v.lineColor == lineColor then
            v.filterFunc = filterFunc
            return
        end
    end
    
    -- Otherwise, insert it
    
    table.insert(self.colorFilters, {lineColor=lineColor, filterFunc=filterFunc})
    
end

function DevConsole:setHighlightFilter(...)
    self.highlightFilter = self:getFilterFunction(...)
end
    
function DevConsole:setDisplayFilter(...)
    self.displayFilter = self:getFilterFunction(...)
end

function DevConsole:setPrintFilter(...)
    self.printFilter = self:getFilterFunction(...)
end

function DevConsole:setMinimized(asMinimized)
    if asMinimized then self:minimize() else self:maximize() end
end

function DevConsole:setMaximized(asMaximized)
    self:setMinimized(not asMaximized)
end

function DevConsole:setMinimizeTop(atTop)
    self.minimizeTop = atTop
    self.minimizeBottom = not self.minimizeTop
end

function DevConsole:setMinimizeBottom(atBottom)
    self:setMinimizeTop(not atBottom)
end

function DevConsole:setMinimizeLeft(atLeft)
    self.minimizeLeft = atLeft
    self.minimizeRight = not self.minimizeLeft
end

function DevConsole:setMinimizeRight(atRight)
    self:setMinimizeLeft(not atRight)
end

function DevConsole:setMinimizeAcross(across)
    self.minimizeAcross = across
    self.minimizeUpDown = not self.minimizeAcross
end

function DevConsole:setMinimizeUpDown(upDown)
    self:setMinimizeAcross(not upDown)
end

function DevConsole:setHidden(asHidden)
    if asHidden then self:hide() else self:show() end
end

function DevConsole:setFontSize(newFontSize)
    self.fontSize = math.max(math.min(newFontSize, DevConsole.MAX_FONT_SIZE),DevConsole.MIN_FONT_SIZE)
end

-- Create getter and setter methods for all console state variables without special behaviors already defined

function DevConsole.createSetters()
    for k,v in pairs(DevConsole) do
        if string.match(k,"^PARM_") and type(v) == "string" then
            local setterName = "set" .. v:gsub("^%l", string.upper)
            if not DevConsole[setterName] then
                DevConsole[setterName] = function(self,value)
                    self[v] = value
                end
            end
            
            local getterName = "get" .. v:gsub("^%l", string.upper)
            if not DevConsole[getterName] then
                DevConsole[getterName] = function(self)
                    return self[v]
                end
            end
            
        end
    end
end
DevConsole.createSetters()

-- Print something to the console using a color given as the first parameter
function DevConsole:printc(...)
    self:printWithColor(...)
end

-- Print something to the console using a comma-space as a separator
function DevConsole:printcs(...)
    self:print(DevConsole.PRINT_COMMA_SPACE_SEPARATOR, ...)
end

-- Print something to the console, expanding tables to key=value pairs
function DevConsole:printt(...)
    self:print(DevConsole.PRINT_EXPAND_TABLES, ...)
end

-- Print something to the console, expanding tables to key=value pairs and using a comma separator
function DevConsole:printtc(...)
    self:print(DevConsole.PRINT_EXPAND_TABLES, DevConsole.PRINT_COMMA_SEPARATOR, ...)
end

-- Print something to the console, expanding tables to key=value pairs and using a comma-space separator
function DevConsole:printtcs(...)
    self:print(DevConsole.PRINT_EXPAND_TABLES, DevConsole.PRINT_COMMA_SPACE_SEPARATOR, ...)
end

-- Print something to the console, expanding tables to key=value pairs and using a newline separator
function DevConsole:printtn(...)
    self:print(DevConsole.PRINT_EXPAND_TABLES, DevConsole.PRINT_NEWLINE_SEPARATOR, ...)
end

-- Print something to the console, using a space separator
function DevConsole:prints(...)
    self:print(DevConsole.PRINT_SPACE_SEPARATOR, ...)
end

-- Print only on a condition
function DevConsole:printIf(condition, ...)
    if condition then self:print(...) end
end
    
-- Print with color only on a condition
function DevConsole:printIfWithColor(condition, withColor, ...)
    if condition then self:printWithColor(withColor, ...) end
end
    
-- Print with color only on a condition
function DevConsole:printIfc(condition, withColor, ...)
    if condition then self:printWithColor(withColor, ...) end
end
    
--[[
function DevConsole:printTable(...)
    local j = select("#", ...)
    local s = ""
    for i=1,j do
        local t = select(i,...)
        if type(t) == "table" then
            for k,v in pairs(t) do
                self:print(s,k,":",v)
            end
            s = ""
        else s = s .. tostring(t)
        end
    end 
end
--]]

-- Print a line using a specific color (pass true to use the default highlight color)
function DevConsole:printWithColor(withColor, ...)
    
    -- Convert the given color
    
    if withColor == true then withColor = self:createPrintColor( self.highlightColor )  
    elseif withColor == nil then withColor = self:createPrintColor( self.textColor )
    elseif type(withColor) == "number" then
        withColor = self:createPrintColor( self.palette[((withColor-1) % #self.palette)+1] )
    elseif type(withColor) == "userdata" and withColor.r and withColor.g and withColor.b and withColor.a then
        withColor = self:createPrintColor( withColor )
    else
        self:printWithColor(true,"WARNING: printWithColor() called with an invalid color parameter")
        --withColor = self.textColor
    end
    
    self:print(withColor, ...)
end 
    
function DevConsole:print(...)
    
    if not self.enabled then return end
    
    -- Build the line to print
    
    local withColor = self.textColor
    local expandTables = false
    local sep = self.printSeparator or ""
    local p = "" -- Paramater passed to print
    local s = "" -- Accumulated string
    local j = select("#", ...)
    for i=1,j do
        p = select(i, ...)
        
        if p == DevConsole.PRINT_HIGHLIGHT then
            withColor = self.highlightColor
        elseif type(p) == "table" and DevConsole.PRINT_SEPARATORS[p] then
            sep = p.printSeparator or self.printSeparator or " "
        elseif p == DevConsole.PRINT_EXPAND_TABLES then
            expandTables = true
        elseif p == DevConsole.PRINT_TEXT_COLOR then
            withColor = self.textColor
        elseif DevConsole.PRINT_COLORS[p] ~= nil and type(p) == "table" then
            if type(p.printColor) == "number" then 
                withColor = self.palette[((p.printColor-1) % #self.palette)+1]
            elseif type(p.printColor) == "userdata" then
                withColor = p.printColor or self.highlightColor
            end
        elseif expandTables and type(p) == "table" then
            if s ~= "" then s = s .. sep end
            if self.numberFormat and self.numberFormat > "" and type(p) == "number" then
                s = s .. string.format(self.numberFormat, p)
            else
                s = s .. tostring(p)
            end
            for k,v in pairs(p) do
                s = s .. sep .. tostring(k) .. "=" .. tostring(v)
            end
        else
            if s ~= "" then s = s .. sep end
            if self.numberFormat and self.numberFormat > "" and type(p) == "number" then
                s = s .. string.format(self.numberFormat, p)
            else
                s = s .. tostring(p)
            end
        end
    end
    
    -- If the console is locked, save the output for later
    
    -- A line has:
    --      str = the text of the line
    --      col = the display text color
    --      lin = the line number
    --      cnt = the count of the number of consecutive prints of this text and color
    --      tim = the time this line was added (OS time, i.e. epoch milliseconds)
    --      lst = the time a duplicate was last added to this line (again, OS time)
    
    local bufferLine = {str=s, col=withColor, lin=self.nextLine, cnt=1,tim=os.time(),lst=os.time()}
    
    -- If there is a print filter, apply it (if it returns false, don't print this line)
    
    if self.printFilter ~= nil then
        if not self.printFilter(bufferLine) then return end
    end
    
    -- If the display is locked, save the line to add to the buffer later and exit here
    
    if self.locked then
        local pendingLine = bufferLine
        table.insert(self.pendingLines, pendingLine)
        return
    end
    
    -- Add the line to the buffer
    -- The priorLine is the last line actually added to the display
    
    -- If the prior line matches the current line, don't add it, just count it
    if self.priorLine and s == self.priorLine.str and withColor == self.priorLine.col then
        self.priorLine.cnt = self.priorLine.cnt + 1 -- Don't duplicate lines, count them
        self.priorLine.lst = os.time() -- But update the last time for the most recent print
    else
        self.priorLine = bufferLine
        table.insert(self.buffer, self.priorLine)
    end
    
    -- Echo
    
    if self.echo then DevConsole.luaPrint(s) end
    
    self:afterPrint()
end

-- Manage print tasks after adding one or more lines to the buffer

function DevConsole:afterPrint()
    
    -- Manage the buffer / limit
    
    self.nextLine = self.nextLine + 1
    if self.bufferLimit > 0 and #self.buffer > self.bufferLimit then
        table.remove(self.buffer, 1)
    end
    
    -- Trigger maximize / show
    
    if self.minimized and self.maxOnPrint then
        self:maximize()
    elseif self.showOnPrint then
        self:show()
    end
end

function DevConsole:zoomIn()
    if self.fontSize >= DevConsole.MAX_FONT_SIZE then
        self.fontSize = DevConsole.MAX_FONT_SIZE
    else
        self.fontSize = self.fontSize + DevConsole.FONT_INCREMENT
    end
end

function DevConsole:zoomOut()
    if self.fontSize <= DevConsole.MIN_FONT_SIZE then
        self.fontSize = DevConsole.MIN_FONT_SIZE
    else
        self.fontSize = self.fontSize - DevConsole.FONT_INCREMENT
    end
end

function DevConsole:zoomReset()
    self.fontSize = self.resetFontSize
end

function DevConsole:minimize()
    local corner
    if self.minimizeRight then
        if self.minimizeTop then 
            if self.minimizeAcross then corner = DevConsole.MINIMIZED.trh else corner = DevConsole.MINIMIZED.trv end
        else
            if self.minimizeAcross then corner = DevConsole.MINIMIZED.brh else corner = DevConsole.MINIMIZED.brv end
        end
    else
        if self.minimizeTop then
            if self.minimizeAcross then corner = DevConsole.MINIMIZED.tlh else corner = DevConsole.MINIMIZED.tlv end
        else
            if self.minimizeAcross then corner = DevConsole.MINIMIZED.blh else corner = DevConsole.MINIMIZED.blv end
        end
    end
    table.insert(corner, self)
    self.minimized = {
        corner=corner,
        where=#corner, 
        label=self.minimizedLabel
    }
    self:computeDimensions()
end

function DevConsole:maximize()
    if not self.minimized then return end
    table.remove(self.minimized.corner, self.minimized.where)
    -- Reset "where" for all other minimized consoles in that set
    for i,console in ipairs(self.minimized.corner) do
        console.minimized.where = i
    end
    self.minimized = nil
    self:show()
end

function DevConsole:show()
    self.hidden = false
end

function DevConsole:hide()
    self.hidden = true
end

function DevConsole:toBack()
    if self:isBack() then return end
    for i,console in ipairs(DevConsole.CONSOLES) do
        if console == self then
            table.remove(DevConsole.CONSOLES, i)
            break
        end
    end
    table.insert(DevConsole.CONSOLES, 1, self)
end

function DevConsole:isBack()
    return self == DevConsole.CONSOLES[1]
end

function DevConsole:toFront()
    if self:isFront() then return end
   for i,console in ipairs(DevConsole.CONSOLES) do
        if console == self then
            table.remove(DevConsole.CONSOLES, i)
            break
        end
    end
    table.insert(DevConsole.CONSOLES, self)
end

function DevConsole:isFront()
    return self == DevConsole.CONSOLES[#DevConsole.CONSOLES]
end

-- Save the current console buffer to a tab
function DevConsole:saveToTab(tabName)
    
    tabName = tabName or self.tabName or ""
    
    -- Test that a tab name exists
    
    if type(self.tabName) ~= "string" or string.len(self.tabName) == 0 then
        self:printWithColor(true, "Invalid tab name.  Cannot save to a project tab.")
        return
    end
    
    -- Make sure the name is valid (replace invalid characters with underscores)
    
    local newTabName = string.gsub(self.tabName,"[^%w]","_")
    
    -- If an empty tab must be used, make a new one
    
    local tabExists = false
    if self.toEmptyTabOnly then
        -- Find and save the latest matching tab name
        local lastName = ""
        for k,tabName in ipairs(listProjectTabs()) do
            if string.match(tabName,"^" .. newTabName) then
                if tabName > lastName then lastName = tabName end
            end
        end
        -- Also look in tabs saved this execution, since Codea seems to be unaware of those
        for k,tabName in ipairs(self.savedTabs) do
            if string.match(tabName,"^" .. newTabName) then
                if tabName > lastName then lastName = tabName end
            end
        end
        local currentData = readProjectTab(lastName)
        -- If last matching tab is not not empty
        if string.len(currentData:gsub("%s","")) > 0 then
            -- If a matching name was
            if lastName > "" and string.match(lastName, "_%d+$") then
                -- If it ends with _%d, replace it with the next number
                newTabName = newTabName .. "_" .. tostring(tonumber(string.match(lastName,"%d+$")) + 1)
            else
                -- Otherwise append _1
                newTabName = newTabName .. "_1"
            end
        end
    else
        for k,tabName in ipairs(listProjectTabs()) do
            if tabName == newTabName then tabExists = true; break end
        end
    end
    
    -- Get the data from the existing tab
    
    local currentData = ""
    if tabExists then
        currentData = readProjectTab(newTabName) or ""
        if self.toEmptyTabOnly and string.len(currentData:gsub("%s","")) > 0 then
            self:printWithColor(true, DevConsole.ICON_ALERT .. " The target tab is not empty.  Cannot (will not) save.")
            return
        end
    end
    
    --  Construct the output
    
    local newData = ""
    if self.appendToTab then newData = currentData .. "\n" end
    newData = newData .. self:getAsString()
    
    -- Do the actual save
    
    saveProjectTab(newTabName, newData)
    
    table.insert(self.savedTabs, newTabName)
       
    -- Log the save action itself
    
    self:printWithColor(true, "Saved to ",newTabName)
end
    
----------------------------------------------------------------------------------------------------------------------------------
-------------------------- Drawing
----------------------------------------------------------------------------------------------------------------------------------

-- Clip to a new clipping rectangle... when unclipping, re-clip using the previous clipping rectangle
function DevConsole:clip(x, y, w, h)
    if not self.clipRects then self.clipRects = {} end
    if x then
        clip(x, y, w, h)
        table.insert(self.clipRects, 1, {x=x,y=y,w=w,h=h})
    else
        if #self.clipRects > 0 then
            table.remove(self.clipRects, 1)
        end
        if #self.clipRects > 0 then
            clip(self.clipRects[1].x, self.clipRects[1].y, self.clipRects[1].w, self.clipRects[1].h)
        else
            clip()
        end
    end
end

function DevConsole:placeButton(button, left, right, top, bottom, seq)
    local bs = DevConsole.BUTTON_SIZE -- Button size
    local bm = DevConsole.BUTTON_SPACE -- Button margin
    if button.atLeft then
        button.x = left  + self.margin + ((seq or 1) - 0.5) * (bs + bm) -- - 0.5 leaves a slight margin for resizing touches
    else
        button.x = right - 2 * self.margin - ((seq or 1) - 0.5) * (bs + bm) -- - 0.5 leaves a slight margin for resizing touches
    end
    if button.atBottom then
        button.y = bottom + 1.5 * self.margin + 5
    else
        button.y = top - 1.5 * self.margin - 5
    end
    button.w = bs
    button.h = bs
end

-- Compute the current dimension information (used by draw, touch)
function DevConsole:computeDimensions()
    self.dimensions = {}
    
    --if self.fullscreen then
    --    self.dimensions = {x=0,y=0,w=WIDTH,h=HEIGHT}
    --else
        -- Compute console x/y/width/height as a fraction of WIDTH/HEIGHT if <= 1
        self.dimensions = {x=self.x, y=self.y, w=self.w, h=self.h} -- Current console x,y,width,height
        if self.x >= 0 and self.x <= 1 then self.dimensions.x = WIDTH * self.x end
        if self.y >= 0 and self.y <= 1 then self.dimensions.y = HEIGHT * self.y end
        if self.w <= 1 then self.dimensions.w = WIDTH * self.w end
        if self.h <= 1 then self.dimensions.h = HEIGHT * self.h end
    --end
    
    if self.minimized then
        
        local where = self.minimized.where
        -- If the horizontal has a spot taken, vertical has to shift one
        if self.minimized.corner == DevConsole.MINIMIZED.trv then
            if #DevConsole.MINIMIZED.trh > 0 then where = where + 1 end
        elseif self.minimized.corner == DevConsole.MINIMIZED.tlv then
            if #DevConsole.MINIMIZED.tlh > 0 then where = where + 1 end
        elseif self.minimized.corner == DevConsole.MINIMIZED.brv then
            if #DevConsole.MINIMIZED.brh > 0 then where = where + 1 end
        elseif self.minimized.corner == DevConsole.MINIMIZED.blv then
            if #DevConsole.MINIMIZED.blh > 0 then where = where + 1 end
        end
        
        if self.minimized.corner == DevConsole.MINIMIZED.trh then
            self.dimensions.x = WIDTH - where * (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = HEIGHT - (1.5 * DevConsole.BUTTON_SIZE)
        elseif self.minimized.corner == DevConsole.MINIMIZED.brh then
            self.dimensions.x = WIDTH - where * (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = 1.5 * DevConsole.BUTTON_SIZE
        elseif self.minimized.corner == DevConsole.MINIMIZED.tlh then
            self.dimensions.x = where * (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = HEIGHT - (1.5 * DevConsole.BUTTON_SIZE)
        elseif self.minimized.corner == DevConsole.MINIMIZED.blh then
            self.dimensions.x = where * (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = 1.5 * DevConsole.BUTTON_SIZE
        elseif self.minimized.corner == DevConsole.MINIMIZED.trv then
            self.dimensions.x = WIDTH - (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = HEIGHT - where * (1.5 * DevConsole.BUTTON_SIZE)
        elseif self.minimized.corner == DevConsole.MINIMIZED.brv then
            self.dimensions.x = WIDTH - (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = where * (1.5 * DevConsole.BUTTON_SIZE)
        elseif self.minimized.corner == DevConsole.MINIMIZED.tlv then
            self.dimensions.x = (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = HEIGHT - where * (1.5 * DevConsole.BUTTON_SIZE)
        elseif self.minimized.corner == DevConsole.MINIMIZED.blv then
            self.dimensions.x = (1.5 * DevConsole.BUTTON_SIZE)
            self.dimensions.y = where * (1.5 * DevConsole.BUTTON_SIZE)
        else
            error("Unknown minimization corner!")
        end
        
        self.dimensions.w = DevConsole.BUTTON_SIZE
        self.dimensions.h = DevConsole.BUTTON_SIZE
        self.minimized.x = self.dimensions.x
        self.minimized.y = self.dimensions.y
        self.minimized.w = self.dimensions.w
        self.minimized.h = self.dimensions.h
    end
    
    
    self.dimensions.vw = self.dimensions.w - self.margin -- Visible console area width
    local headerSize = self.titleSize + self.margin
    self.dimensions.vh = self.dimensions.h - self.margin - headerSize -- Visible console area height = height - bottom margin and top header
    
    self.dimensions.right = self.dimensions.x + self.dimensions.w
    self.dimensions.center = self.dimensions.x + (self.dimensions.w / 2)
    self.dimensions.top = self.dimensions.y + self.dimensions.h 
    
    self.dimensions.ty = self.dimensions.top - headerSize -- Title area vertical base (x)
    
    return  self.dimensions.x,
            self.dimensions.y,
            self.dimensions.w,
            self.dimensions.h,
            self.dimensions.right,
            self.dimensions.center,
            self.dimensions.top,
            self.dimensions.vw,
            self.dimensions.vh
end

function DevConsole:drawButton(button)
    
    font("Arial") -- Font for button labels
    fontSize(button.h * .85) -- Font size for button labels
    strokeWidth(2)
    textMode(CENTER)
    stroke(self.borderColor)
    if button.highlight then fill(self.borderColor) else fill(self.bgColor) end
    
    rect(button.x, button.y, button.w, button.h)
    --self:drawRoundRect(button.x, button.y, button.w, button.h) -- This just doesn't look good with buttons this small
    
    if button.highlight then fill(self.bgColor) else fill(self.textColor) end
    text(button.label, button.x + button.w / 2, button.y + button.h / 2)

end

--[[ This just doesn't look good with buttons this small... the rounded corners are jagged using such a small ellipse()
function DevConsole:drawRoundRect(x, y, w, h, r)
    
    pushStyle()
    
    r = r or math.min(w,h) / 5
    local sw = strokeWidth()
    
    strokeWidth(0)
    
    -- Compute some locations
    
    local lo = x -- Outer Left
    local ro = x + w - 1 -- Outer Right
    local bo = y -- Outer Bottom
    local to = y + h - 1 -- Outer Top
    local wo = w
    local ho = h
    local li = lo - 1 + r -- Inner Left
    local ri = ro - r -- Inner Right
    local bi = bo - 1 + r -- Inner Bottom
    local ti = to - r -- Inner Top
    local wi = w - 2*r -- Inner width (between corners)
    local hi = h - 2*r -- Inner height (between corners)

    -- Draw two rects, to fill everything except the rounded corners NO BORDER
    
    rectMode(CORNER)
    rect(li, bo, wi, ho)
    rect(lo, bi, wo, hi)
    
    -- Draw the straight borders
    
    strokeWidth(sw)
    line( li, to, ri, to ) -- Top
    line( li, bo, ri, bo ) -- Bottom
    line( lo, bi, lo, ti ) -- Left
    line( ro, bi, ro, ti ) -- Right
    
    -- Draw the rounded corners
    ellipseMode(CENTER)
    
    self:clip(lo, bo, r, r)
    ellipse(li, bi, 2*r, 2*r) -- Bottom Left
    self:clip()
    
    self:clip(lo, ti, r, r) -- Top Left
    ellipse(li, ti, 2*r, 2*r)
    self:clip()
    
    self:clip(ri, bo, r, r)
    ellipse(ri, bi, 2*r, 2*r) -- Bottom Right
    self:clip()
    
    self:clip(ri, ti, r, r) -- Top Right
    ellipse(ri, ti, 2*r, 2*r)
    self:clip()
    
    popStyle()
end
--]]

-- Draw this console where it belongs
function DevConsole:draw()
    
    if self.hidden then return end
    
    -- Handle accumulated touches
    
    self:handleTouches()

    -- Local definitions
    
    local x, y = self.margin, self.margin -- Starting x/y locations for buffer contents
    local len = #self.buffer -- Current console buffer size
    local contentSize = vec2() -- Temp total content size accumulator
    local w, h -- Temp width/height variables
    local bufferLine -- The current buffer line
    local str -- Temp string (line of text) variable
    local col -- Temp color variable
    local display -- Temp boolean control to display or skip a line
    local linFmt -- Line number format string
    
    local cx, cy, cw, ch, right, center, top, vw, vh = self:computeDimensions()
    -- cy = cy + self.margin
    local tw, th -- Title width and height
    
    -- Play nice with others
    
    pushStyle()
    pushMatrix()
    
    -- Draw frame and background
    
    textMode(CORNER)
    
    fill(self.bgColor)
    stroke(self.borderColor)
    strokeWidth(self.borderWidth)
    
    if self.minimized then
        self:drawButton(self.minimized)
        popStyle()
        popMatrix()
        return
    end
    
    rect(cx, cy, cw, ch)
    
    -- Draw buttons
    
    for _,atLeft in ipairs{true, false} do
        local offset = 0
        for _,button in ipairs(self.buttons) do
            if button.atLeft == atLeft then
                offset = offset + 1
                self:placeButton(button, cx, right, top, cy, offset)
                self:drawButton(button)
            end
        end 
    end
    -- Right side buttons
    textMode(CORNER)
    
    -- Set font for content
    
    font(self.fontName)
    
    -- Draw title
    
    textWrapWidth(0)
    fontSize(self.titleSize)
    fill(self.titleColor)
    tw,th = textSize(self.title)
    text(self.title, center - (tw / 2), top - self.margin / 2 - th)
    
    -- Clip contents of console in case of overflow
    
    self:clip(cx, cy, vw, vh + self.margin) -- Clip below title
    translate(cx , cy)
    
    fontSize(self.fontSize)
    translate(self.offset.x, self.offset.y) -- Offset contents for scrolling
    
    -- If the content from the last draw has not changed and is smaller than the console, shift everything up
    
    if self.contentSize and self.contentSize.y > 0 then
        if self.contentSize.y < vh then
            translate(0, vh - self.contentSize.y)
        end
    end
    
    if self.wrapLogLines then textWrapWidth(vw-10) end
    
    lineFmt = "%02d  "
    if #self.buffer > 0 then
        linFmt = "%0" .. math.max(2,string.len(tostring(self.buffer[#self.buffer].lin))) .. "d  "
    end
    
    for i=len,1,-1 do
        bufferLine = self.buffer[i]
        if bufferLine and (self.displayFilter == nil or self.displayFilter(bufferLine)) then
                
            str = bufferLine.str
            if self.printLineNumbers then
                str = string.format(linFmt, bufferLine.lin) .. str
            end
            if self.printTime then -- Currently print the most recent log time for repeated lines
                if self.timeFormat == "%f" then
                    str = bufferLine.lst .. " " .. str
                else
                    str = os.date(self.timeFormat, bufferLine.lst) .. " " .. str
                end
            end
            col = bufferLine.col or self.textColor
                
            if self.highlightFilter and self.highlightFilter(bufferLine) then col = self.highlightColor end
            for i,colorFilter in pairs(self.colorFilters) do
                if colorFilter.filterFunc(bufferLine) then 
                    col = colorFilter.lineColor
                    break 
                end    
            end
                
            fill(col)
            text(str, x, y)
            w, h = textSize(str)
                
            if bufferLine.cnt > 1 then
                fill(self.countColor)-- TODO: highlight counts differently
                text(" ✶ " .. tostring(bufferLine.cnt), x + w, y)  -- alternate marker ❉ ✶
            end
                
            if w > contentSize.x then contentSize.x = w end
            contentSize.y = contentSize.y + h + self.lineGap
            y = y + h + self.lineGap
            -- Work on this; requires knowledge of translations like self.offset.y and content-repositioning to top
            --if y + self.offset.y > self.dimensions.top then break end
        end
    end
    
    -- Save the state as things were just drawn
    
    self.contentSize = contentSize
    
    -- Fix the offset, in case the display has been recomputed in a way that makes the offset invalid
    
    if self.offset.x < self.dimensions.vw - contentSize.x then self.offset.x = self.dimensions.vw - contentSize.x end
    if self.offset.x > 0 then self.offset.x = 0 end
    if self.offset.y < self.dimensions.vh - contentSize.y then self.offset.y = self.dimensions.vh - contentSize.y end
    if self.offset.y > 0 then self.offset.y = 0 end
    
    -- Play nice with others
    
    self:clip()
    popMatrix()
    popStyle()
end
    
----------------------------------------------------------------------------------------------------------------------------------
-------------------------- Touch Handling
----------------------------------------------------------------------------------------------------------------------------------

    
-- Capture touches if they are inside this console
function DevConsole:touched(touch)
    
    -- If the touch already belongs to this console, ignore the stuff that short circuits it
    
    if not self.touches[touch.id] then
        
        -- If the console is hidden, it can't possibly accept touches
        
        if self.hidden then return false end
        
        -- Exclude touches not in this console (unless the original touch was in this console to begin with)
        
        if not self.dimensions then self:computeDimensions() end
            
        if touch.x < self.dimensions.x or touch.x > self.dimensions.right
        or touch.y < self.dimensions.y or touch.y > self.dimensions.top then 
            return false
        end
        
    end
        
    -- Add the touch
    
    if touch.state == BEGAN then 
        self.touches[touch.id] = {first=touch, recent=touch, hasMoved = false, lastAction=0} -- lastAction 0 makes sure the first touch is instantaneous
        self.touchCount = self.touchCount + 1
    elseif self.touches[touch.id] ~= nil then
        if touch.state == CANCELLED then
            self.touches[touch.id] = nil
            self.touchCount = self.touchCount - 1
        else
            -- Note: Prior is not tracked here, to make sure prior is always the prior handled touch (in case any are skipped)
            self.touches[touch.id].recent = touch
        end
    end
    
    return true
    
end

-- Handle touches if they are inside this console
function DevConsole:handleTouches()
    
    -- Save time by leaving immediately if no touches to consider
    
    if self.touchCount == 0 then return end -- Optimization, short circuit all this effort if nothing to see here
    
    -- If the console is hidden, it can't possibly accept touches
    
    if self.hidden then
        self.touches = {}
        self.touchCount = 0
        return
    end
    
    -- Prepare to remove cancelled/ended touches, and count touches
    
    local remove = {} -- List of touch sets to remove when done with them
    local maxTapCount = 0
    for id,t in pairs(self.touches) do
        maxTapCount = math.max(maxTapCount, t.first.tapCount)
        if not t.endTime then
            t.endTime = os.clock() + DevConsole.TAP_COUNT_DELAY -- Delay to wait for the next tap
        end
        if t.recent.state == ENDED and t.endTime <= os.clock() -- Only eliminate touches when having waited long enough to see if they were taps
        or t.recent.state == CANCELLED
        or t.ignore then
            remove[id] = t
        end
    end
    
    -- Elminate lesser tap counts
    
    for id,t in pairs(self.touches) do
        if t.first.tapCount < maxTapCount then
            t.ignore = true 
            remove[id] = t
        end
    end
    
    -- Handle all touches
    
    -- Only handle single or two-touch gestures; ignore 3 or more touches at once
    local priorTouchSet
    for id,touchSet in pairs(self.touches) do
        if self.touchCount == 1 then 
            self:handleTouch(touchSet) -- Handle single touch
        elseif priorTouchSet then
            self:handleTwoTouches(priorTouchSet, touchSet) -- Handle
        end
        priorTouchSet = touchSet
    end
    
    -- Actually remove ended/cancelled touches
    
    for id,touchSet in pairs(remove) do 
        if touchSet.button then touchSet.button.highlight = false end
        if self.touches[id] then
            self.touches[id] = nil
        end
    end
    
    -- Reset the touch count
    self.touchCount = 0
    for _,_ in pairs(self.touches) do
        self.touchCount = self.touchCount + 1
    end
end

-- Handle a set of single touches
function DevConsole:handleTouch(touchSet)
    
    -- If this touch set has been handled and cannot act again, just leave (example: minimize button)
    
    if touchSet.ignore then return end
    
    -- Set up what we want to concisely manage this touch
    
    local first = touchSet.first -- Original touch
    local recent = touchSet.recent -- Most recent touch
    local prior = touchSet.prior or first -- Most recent not-ignored touch
    
    -- If minimized, all we can do is unminimize
    
    if self.minimized then 
        self:maximize() 
        return
    end
        
    -- Do buttons first
        
    for i,button in ipairs(self.buttons) do
        if  first.pos.x >= button.x and first.pos.x <= button.x + button.w -- First touch in button
        and first.pos.y >= button.y and first.pos.y <= button.y + button.h then
            local delay = os.clock() - touchSet.lastAction -- Compute the delay since the last action for this touch set (used for continuous zoom buttons)
            if  recent.pos.x >= button.x and recent.pos.x <= button.x + button.w -- New touch also in button
            and recent.pos.y >= button.y and recent.pos.y <= button.y + button.h then
                touchSet.button = button
                button.highlight = true
                if delay >= button.delay then -- delay exceeded 
                    touchSet.prior = recent -- Count this touch as the last one applied for this button
                    touchSet.lastAction = os.clock()
                    if button.delay <= 0 then touchSet.ignore = true end
                    button.action()
                    if touchSet.ignore or touchSet.recent.state == ENDED or touchSet.recent.state == CANCELLED then
                        button.highlight = false
                    end
                    return
                end
            else
                button.highlight = false
            end
            return
        end
    end
    
    -- Remember if this touch set has ever moved (with a slight allowance for an unsteady hand)
    
    touchSet.hasMoved = touchSet.hasMoved or math.abs(first.x - recent.x) > 10 or math.abs(first.y - recent.y) > 10
    
    -- Do long-press title to hide
    
    if  not touchSet.hasMoved
    and recent.timestamp - first.timestamp > 1 -- Hold for one second or more
    and first.y > self.dimensions.ty then -- In title area
        self:hide()
        return
    end
        
    -- Do resizing
    
    --if recent.state == BEGAN and not self.fixedSize then -- determine if in side or bottom margins or corners for resizing
    if not touchSet.resizing and not touchSet.notResizing and not self.fixedSize and touchSet.hasMoved then  
        local resizeMargin = math.max(DevConsole.RESIZE_MARGIN, self.dimensions.top - self.dimensions.ty)
        local isLeft   = first.x < self.dimensions.x + resizeMargin
        local isBottom = first.y < self.dimensions.y + resizeMargin 
        local isRight  = first.x > self.dimensions.right - resizeMargin 
        local isTop    = first.y > self.dimensions.top - resizeMargin
        
        -- Resize only from left, right or bottom, OR upper left or right corners (not top alone, that's for dragging)
        
        if (isTop and (isLeft or isRight)) or (isBottom or isLeft or isRight) then
            touchSet.resizing = {
                isLeft=isLeft, isRight=isRight,isTop=isTop,isBottom=isBottom,
                start = {x=self.dimensions.x, y=self.dimensions.y, w=self.dimensions.w, h=self.dimensions.h}
            }
            self:toFront()
        else
            touchSet.notResizing = true
        end
        
    end
    
    if touchSet.resizing then
        -- Handle resizing computations
        if touchSet.resizing.isLeft then
            local x = touchSet.resizing.start.x - (first.x - recent.x)
            local w = touchSet.resizing.start.w + (first.x - recent.x)
            if self.x >= 0 and self.x <= WIDTH then self.x = x / WIDTH else self.x = x end
            if self.w >= 0 and self.w <= WIDTH then self.w = w / WIDTH else self.w = w end
        elseif touchSet.resizing.isRight then
            local w = touchSet.resizing.start.w + (recent.x - first.x)
            if self.w >= 0 and self.w <= WIDTH then self.w = w / WIDTH else self.w = w end
        end
        if touchSet.resizing.isBottom then
            local y = touchSet.resizing.start.y - (first.y - recent.y)
            local h = touchSet.resizing.start.h + (first.y - recent.y)
            if self.y >= 0 and self.y <= HEIGHT then self.y = y / HEIGHT else self.y = y end
            if self.h >= 0 and self.h <= HEIGHT then self.h = h / HEIGHT else self.h = h end
        elseif touchSet.resizing.isTop then
            local h = touchSet.resizing.start.h + (recent.y - first.y)
            if self.h >= 0 and self.h <= HEIGHT then self.h = h / HEIGHT else self.h = h end
        end
        return
    end
    
    -- Do dragging
    
    if not self.fixedPosition then
        -- Drag if in the title area OR already denoted as dragging
        if touchSet.dragging or first.y > self.dimensions.ty and touchSet.hasMoved then
            -- In the title area, but didn't start in a button or resize area, so drag
            if not touchSet.dragging then
                touchSet.dragging = true
                touchSet.dragStart = vec2(self.dimensions.x, self.dimensions.y)
                self:toFront()
            end
            local x = touchSet.dragStart.x + (recent.x - first.x)
            local y = touchSet.dragStart.y + (recent.y - first.y)
            if self.x >= 0 and self.x <= 1 then self.x = x / WIDTH else self.x = x end
            if self.y >= 0 and self.y <= 1 then self.y = y / HEIGHT else self.y = y end
            return
        end
    end
    
    -- Do taps to adjust "top console"
    
    if  first.tapCount > 0
    and recent.state == ENDED
    and not touchSet.hasMoved
    and os.clock() >= touchSet.endTime then

        if first.tapCount == 1 then
            
            self:toFront()
            
            -- Old method, toggled front/back
            -- if self:isFront() then self:toBack() else self:toFront() end
            
        elseif first.tapCount == 2 then
            
            self:toFront()
            if self.x == 0 and self.y == 0 and self.w == 1 and self.h == 1 and self.restore then
                self.x = self.restore.x
                self.y = self.restore.y
                self.w = self.restore.w
                self.h = self.restore.h
                self.restore = nil
            else
                self.restore = {x=self.x, y=self.y, w=self.w, h=self.h}
                self.x, self.y, self.w, self.h = 0, 0, 1, 1
            end
            --self.fullscreen = not self.fullscreen
            
        elseif first.tapCount == 3 then
            
            self:toBack()
            
        elseif first.tapCount == 4 then
            
            self:printSettings()
            
        end
        
        touchSet.ignore = true -- Ignore all future touches (after this one)
        return
    end
    
    -- Do scrolling
    
    if math.abs(recent.y - first.y) > 20 then
        
        if touchSet.offset == nil then touchSet.offset = vec2(self.offset.x, self.offset.y) end
        
        self.offset.y = touchSet.offset.y + (recent.y - first.y)
        
        if self.offset.y > 0 then self.offset.y = 0 end
        if self.contentSize.y < self.dimensions.vh then self.offset.y = 0
        elseif self.dimensions.vh - self.offset.y - self.margin > self.contentSize.y then
            self.offset.y = self.dimensions.vh - self.contentSize.y - self.margin
        end
        
    end
    
    if not self.wrapLogLines and math.abs(recent.x - first.x) > 20 then
        
        if touchSet.offset == nil then touchSet.offset = vec2(self.offset.x, self.offset.y) end
        
        self.offset.x = touchSet.offset.x + (recent.x - first.x)
        
        if self.offset.x > 0 then self.offset.x = 0 end
        if self.contentSize.x < self.dimensions.vw then self.offset.x = 0
        else
            local maxX = self.dimensions.vw - self.contentSize.x
            if self.offset.x < maxX then self.offset.x = maxX end
        end
        
    end
        
end

-- Handle two touches (pinch to zoom)
function DevConsole:handleTwoTouches(touchSet1, touchSet2)
    
    touchSet1.scaleCount = (touchSet1.scaleCount or 0) + 1
    local prior1  = touchSet1.prior or touchSet1.first  -- Original touch 1
    local recent1 = touchSet1.recent -- Most recent touch 1
    local prior2  = touchSet2.prior or touchSet2.first -- Original touch 2
    local recent2 = touchSet2.recent -- Most recent touch 2
            
    -- Remember the prior touch (for next time)
        
    touchSet1.prior = touchSet1.recent
    touchSet2.prior = touchSet2.recent
    
    -- Compute the factor (font size scale change from original touch to most recent touch)
    
    local factor = recent1.pos:dist(recent2.pos) / prior1.pos:dist(prior2.pos)
    
    -- Pin font size
    
    local newFontSize = math.min(math.max(self.fontSize * factor, DevConsole.MIN_FONT_SIZE),DevConsole.MAX_FONT_SIZE)
    
    -- Reset factor based on pinned font size
    
    factor = newFontSize / self.fontSize
    
    -- Set the new font size
    
    self.fontSize = newFontSize
    
    -- If wrapping lines, no scrolling is allowed (change in font size changes everything)
    
    -- If not wrapping lines, scroll based on the center of the two fingers, BUT scroll incrementally, using only 
    --    the most recent and prior touch (not the first touch, unless it is the prior)
    
    if not self.wrapLogLines then
        
        -- Compute the centers of each pair of touches (those are the "drag points")
        
        local recentC = recent1.pos + (recent1.pos - recent2.pos)/2
        local priorC  = prior1.pos + (prior1.pos - prior2.pos)/2
        
        -- Compute the new offsets from the difference in the two drag points
        
        self.offset.x = self.offset.x + (recentC.x - priorC.x)
        self.offset.y = self.offset.y + (recentC.y - priorC.y)
        
        -- Constrain the new offset to avoid coloring outside the lines (ignore the change in content size for now)
        
        if self.offset.x < self.dimensions.vw -  self.contentSize.x then self.offset.x = self.dimensions.vw - self.contentSize.x end
        if self.offset.x > 0 then self.offset.x = 0 end
        if self.offset.y < self.dimensions.vh - self.contentSize.y then self.offset.y = self.dimensions.vh - self.contentSize.y end
        if self.offset.y > 0 then self.offset.y = 0 end
        
    end
end
